/*
 * Copyright (C) 2019 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.apksig.util;

import java.io.RandomAccessFile;

enum DataSourceFromRAFFactory {
    RANDOM_ACCESS_FILE {
        @Override DataSource create(RandomAccessFile file) {
            return DataSources.asDataSource(file);
        }

        @Override DataSource create(RandomAccessFile file, long offset, long size) {
            return DataSources.asDataSource(file, offset, size);
        }
    },
    FILE_CHANNEL {
        @Override DataSource create(RandomAccessFile file) {
            return DataSources.asDataSource(file.getChannel());
        }

        @Override DataSource create(RandomAccessFile file, long offset, long size) {
            return DataSources.asDataSource(file.getChannel(), offset, size);
        }
    };

    abstract DataSource create(RandomAccessFile file);
    abstract DataSource create(RandomAccessFile file, long offset, long size);
}
